# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import sys

from gallia.command.base import BaseCommand
from gallia.commands.discover.doip import DoIPDiscoverer
from gallia.commands.discover.hsfz import HSFZDiscoverer
from gallia.commands.primitive.generic.pdu import GenericPDUPrimitive
from gallia.commands.primitive.uds.dtc import (
    ClearDTCPrimitive,
    ControlDTCPrimitive,
    ReadDTCPrimitive,
)
from gallia.commands.primitive.uds.ecu_reset import ECUResetPrimitive
from gallia.commands.primitive.uds.iocbi import IOCBIPrimitive
from gallia.commands.primitive.uds.pdu import SendPDUPrimitive
from gallia.commands.primitive.uds.ping import PingPrimitive
from gallia.commands.primitive.uds.rdbi import ReadByIdentifierPrimitive
from gallia.commands.primitive.uds.rmba import RMBAPrimitive
from gallia.commands.primitive.uds.rtcl import RTCLPrimitive
from gallia.commands.primitive.uds.vin import VINPrimitive
from gallia.commands.primitive.uds.wdbi import WriteByIdentifierPrimitive
from gallia.commands.primitive.uds.wmba import WMBAPrimitive
from gallia.commands.scan.uds.identifiers import ScanIdentifiers
from gallia.commands.scan.uds.memory import MemoryFunctionsScanner
from gallia.commands.scan.uds.reset import ResetScanner
from gallia.commands.scan.uds.sa_dump_seeds import SASeedsDumper
from gallia.commands.scan.uds.services import ServicesScanner
from gallia.commands.scan.uds.sessions import SessionsScanner

registry: list[type[BaseCommand]] = [
    DoIPDiscoverer,
    ECUResetPrimitive,
    GenericPDUPrimitive,
    HSFZDiscoverer,
    IOCBIPrimitive,
    MemoryFunctionsScanner,
    PingPrimitive,
    RMBAPrimitive,
    RTCLPrimitive,
    ReadByIdentifierPrimitive,
    ResetScanner,
    SASeedsDumper,
    ScanIdentifiers,
    SessionsScanner,
    ServicesScanner,
    ClearDTCPrimitive,
    ControlDTCPrimitive,
    ReadDTCPrimitive,
    ECUResetPrimitive,
    VINPrimitive,
    IOCBIPrimitive,
    PingPrimitive,
    RMBAPrimitive,
    RTCLPrimitive,
    GenericPDUPrimitive,
    SendPDUPrimitive,
    ServicesScanner,
    SessionsScanner,
    VINPrimitive,
    WMBAPrimitive,
    WriteByIdentifierPrimitive,
]

# TODO: Investigate why linters didn't catch faulty strings in here.
__all__ = [
    "DoIPDiscoverer",
    "ECUResetPrimitive",
    "GenericPDUPrimitive",
    "HSFZDiscoverer",
    "IOCBIPrimitive",
    "MemoryFunctionsScanner",
    "PingPrimitive",
    "RMBAPrimitive",
    "RTCLPrimitive",
    "ReadByIdentifierPrimitive",
    "ResetScanner",
    "SASeedsDumper",
    "ScanIdentifiers",
    "SessionsScanner",
    "ServicesScanner",
    "ClearDTCPrimitive",
    "ControlDTCPrimitive",
    "ReadDTCPrimitive",
    "ECUResetPrimitive",
    "VINPrimitive",
    "IOCBIPrimitive",
    "PingPrimitive",
    "RMBAPrimitive",
    "RTCLPrimitive",
    "GenericPDUPrimitive",
    "SendPDUPrimitive",
    "ServicesScanner",
    "SessionsScanner",
    "VINPrimitive",
    "WMBAPrimitive",
    "WriteByIdentifierPrimitive",
]


if sys.platform.startswith("linux"):
    from gallia.commands.discover.find_xcp import CanFindXCP, TcpFindXCP, UdpFindXCP
    from gallia.commands.discover.uds.isotp import IsotpDiscoverer
    from gallia.commands.fuzz.uds.pdu import PDUFuzzer
    from gallia.commands.primitive.uds.xcp import SimpleTestXCP
    from gallia.commands.script.vecu import DbVirtualECU, RngVirtualECU

    registry += [
        CanFindXCP,
        UdpFindXCP,
        TcpFindXCP,
        IsotpDiscoverer,
        PDUFuzzer,
        SimpleTestXCP,
        DbVirtualECU,
        RngVirtualECU,
    ]

    __all__ += [
        "CanFindXCP",
        "UDSFindXCP",
        "TCPFindXCP",
        "IsotpDiscoverer",
        "PDUFuzzer",
        "SimpleTestXCP",
        "DbVirtualECU",
        "RngVirtualECU",
    ]


if sys.platform == "win32":
    from gallia.commands.script.flexray import FRConfigDump, FRDump

    registry += [FRDump, FRConfigDump]
    __all__ += ["FRDump", "FRConfigDump"]



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Namespaces Utility Functions for Declarative Typed Argument Parsing.

The `namespaces` module contains a utility function used for recursively
converting `argparse.Namespace`s to regular Python `dict`s.
"""

import argparse
from typing import Any


def to_dict(namespace: argparse.Namespace) -> dict[str, Any]:
    """Converts a nested namespace to a dictionary recursively.

    Args:
        namespace (argparse.Namespace): Namespace object to convert.

    Returns:
        Dict[str, Any]: Nested dictionary generated from namespace.
    """
    # Get Dictionary from Namespace Vars
    dictionary = dict(vars(namespace))

    # Loop Through Dictionary
    for key, value in dictionary.items():
        # Check for Namespace Objects
        if isinstance(value, argparse.Namespace):
            # Recurse
            dictionary[key] = to_dict(value)

    return dictionary



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import ctypes
import math
import sys
from collections.abc import Callable
from typing import Self, cast

assert sys.platform == "win32", "unsupported platform"

from gallia.log import get_logger
from gallia.transports import _ctypes_vector_xl
from gallia.utils import handle_task_error, set_task_handler_ctx_variable

logger = get_logger(__name__)

WaitForSingleObject: Callable[[int, int], int]
INFINITE: int
try:
    # Try builtin Python 3 Windows API
    from _winapi import WaitForSingleObject
except ImportError:
    raise RuntimeError("platform does not provide WaitForSingleObject")


class FlexRayCtypesBackend:
    def __init__(self, channel_mask: int, port_handle: int) -> None:
        self.channel_mask = channel_mask
        self.port_handle = port_handle
        self.queue: asyncio.Queue[_ctypes_vector_xl.XL_FR_RX_FRAME_EV] = asyncio.Queue()
        self.background_task: asyncio.Task[None] | None = None
        self.event_handle: _ctypes_vector_xl.XLhandle | None = None

    @classmethod
    def create(
        cls,
        channel_no: int | None,
        rx_queue_size: int = 0x2000,
    ) -> Self:
        cls._enable_vector_license()
        cls._open_driver()

        channel_mask = cls._find_flexray_channel(channel_no)
        port_handle = cls._open_flexray_port(
            cls.__class__.__name__,
            channel_mask,
            rx_queue_size,
        )

        return cls(
            channel_mask=channel_mask,
            port_handle=port_handle,
        )

    def set_configuration(self, config: _ctypes_vector_xl.XLfrClusterConfig) -> None:
        _ctypes_vector_xl.xlFrSetConfiguration(
            _ctypes_vector_xl.XLportHandle(self.port_handle),
            _ctypes_vector_xl.XLaccess(self.channel_mask),
            ctypes.byref(config),
        )

    def get_configuration(self) -> _ctypes_vector_xl.XLfrClusterConfig:
        config = _ctypes_vector_xl.XLfrChannelConfig()
        _ctypes_vector_xl.xlFrGetChannelConfiguration(
            _ctypes_vector_xl.XLportHandle(self.port_handle),
            _ctypes_vector_xl.XLaccess(self.channel_mask),
            ctypes.byref(config),
        )
        return cast(_ctypes_vector_xl.XLfrClusterConfig, config.xlFrClusterConfig)

    @staticmethod
    def _open_driver() -> None:
        _ctypes_vector_xl.xlOpenDriver()

    @staticmethod
    def _close_driver() -> None:
        _ctypes_vector_xl.xlCloseDriver()

    @classmethod
    def _find_flexray_channel(cls, no: int | None = None) -> int:
        driver_config = cls.get_xl_driver_config()

        # If no is None, then use the first one, otherwise count.
        for i in range(driver_config.channelCount):
            xlcc: _ctypes_vector_xl.XLchannelConfig = driver_config.channel[i]
            bus_caps = _ctypes_vector_xl.XL_BusCapabilities(xlcc.channelBusCapabilities)
            if bus_caps & _ctypes_vector_xl.XL_BusCapabilities.XL_BUS_ACTIVE_CAP_FLEXRAY:
                if no is not None:
                    if i == no:
                        return _ctypes_vector_xl.XLaccess(xlcc.channelMask).value
                else:
                    return _ctypes_vector_xl.XLaccess(xlcc.channelMask).value

        raise ValueError("no flexray channel found")

    @staticmethod
    def _enable_vector_license() -> None:
        out = ctypes.c_uint()
        _ctypes_vector_xl.xlGetKeymanBoxes(ctypes.byref(out))

    @staticmethod
    def _open_flexray_port(user_name: str, channel_mask: int, rx_queue_size: int) -> int:
        if not math.log2(rx_queue_size).is_integer():
            raise ValueError("rx_queue_size must be a power of 2")
        if not (rx_queue_size >= 8192 and rx_queue_size <= 1048576):
            raise ValueError("rx_queue_size must be within 8192…1048576 bytes (1 MB)")

        port_handle = _ctypes_vector_xl.XLportHandle()
        permission_mask = _ctypes_vector_xl.XLaccess()

        _ctypes_vector_xl.xlOpenPort(
            ctypes.byref(port_handle),
            ctypes.create_string_buffer(user_name.encode("ascii")),
            _ctypes_vector_xl.XLaccess(channel_mask),
            ctypes.byref(permission_mask),
            ctypes.c_uint(rx_queue_size),
            _ctypes_vector_xl.XL_InterfaceVersion.XL_INTERFACE_VERSION_V4,
            _ctypes_vector_xl.XL_BusTypes.XL_BUS_TYPE_FLEXRAY,
        )

        return int(port_handle.value)

    def _close_port(self) -> None:
        _ctypes_vector_xl.xlClosePort(_ctypes_vector_xl.XLportHandle(self.port_handle))

    def _set_notification(self) -> _ctypes_vector_xl.XLhandle:
        event_handle = _ctypes_vector_xl.XLhandle()
        _ctypes_vector_xl.xlSetNotification(
            self.port_handle,
            event_handle,
            1,
        )
        return event_handle

    @staticmethod
    def get_xl_driver_config() -> _ctypes_vector_xl.XLdriverConfig:
        driver_config = _ctypes_vector_xl.XLdriverConfig()
        _ctypes_vector_xl.xlGetDriverConfig(driver_config)
        return driver_config

    def activate_channel(self) -> None:
        _ctypes_vector_xl.xlActivateChannel(
            _ctypes_vector_xl.XLportHandle(self.port_handle),
            _ctypes_vector_xl.XLaccess(self.channel_mask),
            _ctypes_vector_xl.XL_BusTypes.XL_BUS_TYPE_FLEXRAY,
            _ctypes_vector_xl.XL_ACTIVATE_RESET_CLOCK,
        )

    def _deactivate_channel(self) -> None:
        _ctypes_vector_xl.xlDeactivateChannel(
            _ctypes_vector_xl.XLportHandle(self.port_handle),
            _ctypes_vector_xl.XLaccess(self.channel_mask),
        )

    def add_block_all_filter(self) -> None:
        filter_ = _ctypes_vector_xl.XLfrAcceptanceFilter(
            _ctypes_vector_xl.XL_FR_FILTER_BLOCK,
            _ctypes_vector_xl.XL_FR_FILTER_TYPE_DATA
            | _ctypes_vector_xl.XL_FR_FILTER_TYPE_NF
            | _ctypes_vector_xl.XL_FR_FILTER_TYPE_FILLUP_NF,
            ctypes.c_uint(1),
            ctypes.c_uint(255),
            ctypes.c_uint(self.channel_mask),
        )

        _ctypes_vector_xl.xlFrSetAcceptanceFilter(
            _ctypes_vector_xl.XLportHandle(self.port_handle),
            _ctypes_vector_xl.XLaccess(self.channel_mask),
            ctypes.byref(filter_),
        )

    def set_acceptance_filter(
        self,
        from_slot: int,
        to_slot: int,
    ) -> None:
        filter_ = _ctypes_vector_xl.XLfrAcceptanceFilter(
            _ctypes_vector_xl.XL_FR_FILTER_PASS,
            _ctypes_vector_xl.XL_FR_FILTER_TYPE_DATA,
            ctypes.c_uint(from_slot),
            ctypes.c_uint(to_slot),
            ctypes.c_uint(self.channel_mask),
        )

        _ctypes_vector_xl.xlFrSetAcceptanceFilter(
            _ctypes_vector_xl.XLportHandle(self.port_handle),
            _ctypes_vector_xl.XLaccess(self.channel_mask),
            ctypes.byref(filter_),
        )

    def start_queue(self) -> None:
        self.event_handle = self._set_notification()
        self.background_task = asyncio.create_task(self._receive_worker())
        self.background_task.add_done_callback(
            handle_task_error,
            context=set_task_handler_ctx_variable(__name__),
        )

    async def stop_queue(self) -> None:
        assert self.background_task is not None

        self.background_task.cancel()
        await self.background_task

    async def _read_frame(self) -> None:
        while True:
            # Provide a context switch possibility.
            await asyncio.sleep(0)

            event = _ctypes_vector_xl.XLfrEvent()

            try:
                logger.trace("reading from XL queue…")
                _ctypes_vector_xl.xlFrReceive(
                    _ctypes_vector_xl.XLportHandle(self.port_handle),
                    ctypes.byref(event),
                )
            except _ctypes_vector_xl.VectorQueueIsEmptyError:
                logger.trace("XL queue is empty")
                break
            except _ctypes_vector_xl.VectorQueueIsFullError:
                logger.error("receive queue is full, gallia is too slow")
                logger.warn("flushing queue, packages will be dropped")
                _ctypes_vector_xl.xlFlushReceiveQueue(
                    _ctypes_vector_xl.XLportHandle(self.port_handle),
                )
                break

            if event.tag != _ctypes_vector_xl.XL_FR_RX_FRAME:
                continue
            if event.tagData.frRxFrame.payloadLength == 0:
                continue

            logger.trace("got raw event from XL: %s", event.tagData.frRxFrame)
            logger.trace("submitting fr event to queue…")
            await self.queue.put(cast(_ctypes_vector_xl.XL_FR_RX_FRAME_EV, event.tagData.frRxFrame))

    async def _receive_worker(self) -> None:
        assert self.event_handle is not None
        assert self.event_handle.value is not None

        logger.debug("_receive_worker() started")

        try:
            while True:
                poll_intervall = 100

                logger.trace("poller waiting…")
                await asyncio.to_thread(
                    WaitForSingleObject,
                    self.event_handle.value,
                    poll_intervall,
                )

                await self._read_frame()
        except asyncio.CancelledError:
            logger.debug("read worker cancelled")

    async def receive(
        self,
        slot_id: int | None,
        timeout: float | None,
    ) -> _ctypes_vector_xl.XL_FR_RX_FRAME_EV:
        async with asyncio.timeout(timeout):
            while True:
                event = await self.queue.get()
                received_slot_id = event.slotID

                if slot_id is not None:
                    if received_slot_id != slot_id:
                        continue
                return event

    async def transmit(self, slot_id: int, data: bytes) -> None:
        event = _ctypes_vector_xl.XLfrEvent()
        event.tag = _ctypes_vector_xl.XL_FR_TX_FRAME
        event.flagsChip = _ctypes_vector_xl.XL_FR_CHANNEL_A
        event.size = 0  # calculated inside XL-API DLL
        event.userHandle = 0

        event.tagData.frTxFrame.flags = 0x08  # TODO: what is 0x08?
        event.tagData.frTxFrame.offset = 0
        event.tagData.frTxFrame.repetition = 1

        to_send = bytearray(data.ljust(_ctypes_vector_xl.XL_FR_MAX_DATA_LENGTH, b"\x00"))
        # TODO: why is the 0x80 needed???????????
        to_send[73] = 0x80

        # TODO: is this somthing that needs to be configured??
        event.tagData.frTxFrame.payloadLength = 48
        event.tagData.frTxFrame.slotID = slot_id
        event.tagData.frTxFrame.txMode = _ctypes_vector_xl.XL_FR_TX_MODE_SINGLE_SHOT
        event.tagData.frTxFrame.incrementOffset = 0
        event.tagData.frTxFrame.incrementSize = 0

        if len(to_send) > _ctypes_vector_xl.XL_FR_MAX_DATA_LENGTH:
            raise ValueError("frame exceeds max data length")

        event.tagData.frTxFrame.data = (
            ctypes.c_ubyte * _ctypes_vector_xl.XL_FR_MAX_DATA_LENGTH
        ).from_buffer_copy(to_send)

        await asyncio.to_thread(
            _ctypes_vector_xl.xlFrTransmit,
            self.port_handle,
            self.channel_mask,
            ctypes.byref(event),
        )

    async def close(self) -> None:
        await self.stop_queue()
        self._deactivate_channel()
        self._close_port()
        self._close_driver()



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import binascii
import os
import tomllib
from abc import ABC
from collections.abc import Callable
from enum import Enum
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    TypeAlias,
    TypeVar,
    Unpack,
    get_args,
)

from pydantic import BeforeValidator, PlainSerializer
from pydantic.fields import _FromFieldInfoInputs
from pydantic_core import PydanticUndefined

from gallia.config import Config
from gallia.pydantic_argparse import BaseCommand
from gallia.pydantic_argparse.utils.field import ArgFieldInfo
from gallia.utils import unravel, unravel_2d


def err_int(x: str, base: int) -> int:
    try:
        return int(x, base)
    except ValueError:
        base_suffix = ""

        if base != 0:
            base_suffix = f" with base {base}"

        raise ValueError(
            f"{repr(x)} is not a valid representation for an integer{base_suffix}"
        ) from None


AutoInt = Annotated[int, BeforeValidator(lambda x: x if isinstance(x, int) else err_int(x, 0))]
"""
Special type for a field, which automatically parses int from a string.

See int() with base=0 for more information on the syntax.

Usage: x: HexInt = ....
"""


HexInt = Annotated[int, BeforeValidator(lambda x: x if isinstance(x, int) else err_int(x, 16))]
"""
Special type for a field, which parses int from a hex string.

See int() with base=16 for more information on the syntax.

Usage: x: HexInt = ....
"""

HexBytes = Annotated[
    bytes,
    BeforeValidator(lambda x: x if isinstance(x, bytes) else binascii.unhexlify(x)),
    PlainSerializer(lambda x: binascii.hexlify(x)),
]
"""
Special type for a field, which parses bytes from hex strings.

See binascii.unhexlify() for more information on the syntax.

Usage: x: HexBytes = ....
"""


def _process_ranges(value: Any) -> Any:
    if isinstance(value, str):
        return unravel(",".join(value.split()))
    elif isinstance(value, list) and all(isinstance(x, str) for x in value):
        return unravel(",".join(value))
    return value


Ranges = Annotated[list[int], BeforeValidator(_process_ranges)]
"""
Special type for a field, which parses one-dimensional ranges.

See unravel() for more information on the syntax.

Usage: x: Ranges = ....
"""

Ranges2D = Annotated[
    dict[int, list[int] | None],
    BeforeValidator(
        lambda x: x
        if isinstance(x, dict)
        else unravel_2d(" ".join(x))
        if isinstance(x, list)
        else unravel_2d(x)
    ),
]
"""
Special type for a field, which parses two-dimensional ranges.

See unravel_2d() for more information on the syntax.

Usage: x: Ranges2D = ....
"""


T = TypeVar("T")
EnumType = TypeVar("EnumType", bound=Enum)
LiteralType = TypeVar("LiteralType")


def auto_enum(x: str, enum_type: type[EnumType]) -> EnumType:
    try:
        return enum_type[x]
    except KeyError:
        try:
            return enum_type(x)
        except ValueError:
            try:
                return enum_type(int(x, 0))
            except ValueError:
                pass

    raise ValueError(f"{x} is not a valid key or value for {enum_type}")


if TYPE_CHECKING:
    Idempotent: TypeAlias = Annotated[T, ""]
    EnumArg: TypeAlias = Annotated[EnumType, ""]
    AutoLiteral: TypeAlias = Annotated[LiteralType, ""]
else:

    class _TrickType:
        def __init__(self, function: Callable[[type[T]], type[T]]):
            self.function = function

        def __getitem__(self, cls: type[T]) -> type[T]:
            return self.function(cls)

    Idempotent = _TrickType(
        lambda cls: Annotated[cls, BeforeValidator(lambda x: x if isinstance(x, cls) else cls(x))]
    )
    """
    Wrapper for fields of types which can be instantiated by a certain value, but not by an instance of its own type.

    This way, it is possible to fill the corresponding parameter of the model with both the value as well as an already instantiated object of that class.

    Usage: x: Idempotent[SomeClass] = ...
    """

    EnumArg = _TrickType(
        lambda cls: Annotated[
            cls, BeforeValidator(lambda x: x if isinstance(x, cls) else auto_enum(x, cls))
        ]
    )
    """
    Wrapper for enum fields to provide automatic parsing of enum values by either name or value similar to AutoInt.

    Usage: x: EnumArg[SomeEnum] = ...
    """

    def auto_literal(cls: type[T]):
        args = get_args(cls)

        mapping = {}

        for arg in args:
            if isinstance(arg, Enum):
                mapping[arg.value] = arg
                mapping[arg.name] = arg
            elif isinstance(arg, bytes):
                mapping[binascii.hexlify(arg)] = arg

        def try_auto_literal(value: Any):
            if value in args:
                return value

            try:
                return mapping[value]
            except KeyError:
                pass

            try:
                return mapping[err_int(value, 0)]
            except (KeyError, ValueError):
                pass

            return value

        return Annotated[cls, BeforeValidator(try_auto_literal)]

    AutoLiteral = _TrickType(auto_literal)
    """
    Wrapper for Literal fields to provide automatic handling of enum, int and bytes parsing for values defined in Literals similar to EnumArg, AutoInt and HexBytes.

    Usage: x: AutoLiteral[Literal[1, 2, 3]] = ...
    """


class ConfigArgFieldInfo(ArgFieldInfo):
    def __init__(
        self,
        default: Any,
        positional: bool,
        short: str | None,
        metavar: str | None,
        cli_group: str | None,
        const: Any,
        hidden: bool,
        config_section: str | None,
        **kwargs: Unpack[_FromFieldInfoInputs],
    ):
        """
        Creates a new ConfigArgFieldInfo.

        This is a special variant of pydantic's FieldInfo, which adds several arguments,
        mainly related to CLI arguments and config sections.
        For general usage and details on the generic parameters see https://docs.pydantic.dev/latest/concepts/fields.
        Just as with pydantic's FieldInfo, this should usually not be called directly.
        Instead use the Field() function of this module.

        :param default: The default value, if non is given explicitly.
        :param positional: Specifies, if the argument is shown as positional, as opposed to optional (default), on the CLI.
        :param short: An optional alternative name for the CLI, which is auto-prefixed with "-".
        :param metavar: The type hint which is shown on the CLI for an argument. If none is specified, it is automatically inferred from its type.
        :param cli_group: The group in the CLI under which the argument is listed. If none is specified, it is automatically set to the CLI group of the config class to which this argument belongs.
        :param const: Specifies, a default value, if the argument is set with no explicit value.
        :param hidden: Specifies, that the argument is part of neither the CLI nor the config file.
        :param config_section: Specifies the config section under which the argument is listed. If none is specified, it is automatically set to the config section of the config class to which this argument belongs.
        :param kwargs: Generic pydantic Field() arguments (see https://docs.pydantic.dev/latest/api/fields/#pydantic.fields.FieldInfo).
        """
        super().__init__(
            default=default,
            positional=positional,
            short=short,
            metavar=metavar,
            cli_group=cli_group,
            const=const,
            hidden=hidden,
            **kwargs,
        )

        self.config_section = config_section


def Field(
    default: Any = PydanticUndefined,
    positional: bool = False,
    short: str | None = None,
    metavar: str | None = None,
    cli_group: str | None = None,
    const: Any = PydanticUndefined,
    hidden: bool = False,
    config_section: str | None = None,
    **kwargs: Unpack[_FromFieldInfoInputs],
) -> Any:
    """
    Creates a new ConfigArgFieldInfo.

    This is a special variant of pydantic's Field() function, which adds several arguments,
    mainly related to CLI arguments and config sections.
    For general usage and details on the generic parameters see https://docs.pydantic.dev/latest/concepts/fields.

    :param default: The default value, if non is given explicitly.
    :param positional: Specifies, if the argument is shown as positional, as opposed to optional (default), on the CLI.
    :param short: An optional alternative name for the CLI, which is auto-prefixed with "-".
    :param metavar: The type hint which is shown on the CLI for an argument. If none is specified, it is automatically inferred from its type.
    :param cli_group: The group in the CLI under which the argument is listed. If none is specified, it is automatically set to the CLI group of the config class to which this argument belongs.
    :param const: Specifies, a default value, if the argument is set with no explicit value.
    :param hidden: Specifies, that the argument is part of neither the CLI nor the config file.
    :param config_section: Specifies the config section under which the argument is listed. If none is specified, it is automatically set to the config section of the config class to which this argument belongs.
    :param kwargs: Generic pydantic Field() arguments (see https://docs.pydantic.dev/latest/api/fields/#pydantic.fields.Field).
    :return: A ConfigArgFieldInfo.
    """
    return ConfigArgFieldInfo(
        default, positional, short, metavar, cli_group, const, hidden, config_section, **kwargs
    )


class GalliaBaseModel(BaseCommand, ABC):
    """
    Base class for config classes for commands.


    """

    init_kwargs: dict[str, Any] | None = Field(
        None,
        hidden=True,
        description="This allows to initialize parts or all of the fields safely. Required args may be specified explicitly to please the linter",
    )
    _cli_group: str | None = None
    _config_section: str | None = None
    __config_registry: dict[str, tuple[str, Any]] = {}

    def __init__(self, **data: Any):
        init_kwargs = data.pop("init_kwargs", {})

        if init_kwargs is None:
            init_kwargs = {}
        else:
            # Copy to avoid side effects when reusing init_kwargs
            init_kwargs = dict(init_kwargs)

        init_kwargs.update(data)

        super().__init__(**init_kwargs)

    def __init_subclass__(
        cls,
        /,
        cli_group: str | None = None,
        config_section: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init_subclass__(**kwargs)

        cls._config_section = config_section
        cls._cli_group = cli_group

        for attribute, info in vars(cls).items():
            # Attribute specific annotation takes precedence
            if isinstance(info, ArgFieldInfo) and info.group is None:
                info.group = cli_group

            if isinstance(info, ConfigArgFieldInfo):
                # Attribute specific annotation takes precedence
                if info.config_section is None:
                    info.config_section = config_section

                # Add config to registry
                if info.config_section is not None:
                    config_attribute = (
                        f"{info.config_section}.{attribute}"
                        if info.config_section != ""
                        else attribute
                    )
                    description = "" if info.description is None else info.description

                    # TODO: Private attributes are write protected and throw a TypeError
                    # This is a quick hack to make it work nonetheless
                    try:
                        GalliaBaseModel.__config_registry[config_attribute] = (
                            description,
                            info.default,
                        )
                    except TypeError:
                        GalliaBaseModel.__config_registry = {}

                    GalliaBaseModel.__config_registry[config_attribute] = (
                        description,
                        info.default,
                    )

    @staticmethod
    def registry() -> dict[str, tuple[str, Any]]:
        return GalliaBaseModel.__config_registry

    @classmethod
    def attributes_from_toml(cls, path: Path) -> dict[str, Any]:
        toml_config = tomllib.loads(path.read_text())
        return cls.attributes_from_config(Config(toml_config))

    @classmethod
    def attributes_from_config(cls, config: Config, source: str = "config file") -> dict[str, Any]:
        result = {}

        for name, info in cls.model_fields.items():
            if isinstance(info, ConfigArgFieldInfo):
                config_attribute = (
                    f"{info.config_section}.{name}" if info.config_section != "" else name
                )

                if (value := config.get_value(config_attribute)) is not None:
                    result[name] = (f"{source} ({info.config_section}:{name})", value)

        return result

    @classmethod
    def attributes_from_env(cls) -> dict[str, Any]:
        result = {}

        for name, info in cls.model_fields.items():
            if isinstance(info, ConfigArgFieldInfo):
                config_attribute = f"GALLIA_{name.upper()}"

                if (value := os.getenv(config_attribute)) is not None:
                    result[name] = (f"environment variable ({config_attribute})", value)

        return result



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import ipaddress
import subprocess
from urllib.parse import urlparse

import pydantic
from pydantic.networks import IPvAnyAddress

from gallia.log import get_logger
from gallia.utils import supports_platform

logger = get_logger(__name__)


def split_host_port(
    hostport: str,
    default_port: int | None = None,
) -> tuple[str, int | None]:
    """Splits a combination of ip address/hostname + port into hostname/ip address
    and port.  The default_port argument can be used to return a port if it is
    absent in the hostport argument."""
    # Special case: If hostport is an ipv6 then the urlparser does some weird
    # things with the colons and tries to parse ports. Catch this case early.
    host = ""
    port = default_port
    try:
        # If hostport is a valid ip address (v4 or v6) there
        # is no port included
        host = str(ipaddress.ip_address(hostport))
    except ValueError:
        pass

    # Only parse if hostport is not a valid ip address.
    if host == "":
        # urlparse() and urlsplit() insists on absolute URLs starting with "//".
        url = urlparse(f"//{hostport}")
        host = url.hostname if url.hostname else url.netloc
        port = url.port if url.port else default_port
    return host, port


def join_host_port(host: str, port: int) -> str:
    if ":" in host:
        return f"[{host}]:port"
    return f"{host}:{port}"


class AddrInfo(pydantic.BaseModel):
    family: str
    local: IPvAnyAddress
    prefixlen: int
    broadcast: IPvAnyAddress | None = None
    scope: str
    label: str | None = None
    valid_life_time: int
    preferred_life_time: int

    def is_v4(self) -> bool:
        return self.family == "inet"


class Interface(pydantic.BaseModel):
    ifindex: int
    ifname: str
    flags: list[str]
    mtu: int
    qdisc: str
    operstate: str
    group: str
    link_type: str
    address: str | None = None
    broadcast: str | None = None
    addr_info: list[AddrInfo]

    def is_up(self) -> bool:
        return self.operstate == "UP"

    def can_broadcast(self) -> bool:
        return "BROADCAST" in self.flags


@supports_platform("linux")
def net_if_addrs() -> list[Interface]:
    try:
        p = subprocess.run(["ip", "-j", "address", "show"], capture_output=True, check=True)
    except FileNotFoundError as e:
        logger.warning(f"Could not query information about interfaces: {e}")
        return []

    try:
        iface_list_type = pydantic.TypeAdapter(list[Interface])
        return iface_list_type.validate_json(p.stdout.decode())
    except pydantic.ValidationError as e:
        logger.error("BUG: A special case for `ip -j address show` is not handled!")
        logger.error("Please report a bug including the following json string.")
        logger.error("https://github.com/Fraunhofer-AISEC/gallia/issues")
        logger.error(e.json())
        raise


@supports_platform("linux")
def net_if_broadcast_addrs() -> list[AddrInfo]:
    out = []
    for iface in net_if_addrs():
        if not (iface.is_up() and iface.can_broadcast()):
            continue

        for addr in iface.addr_info:
            if not addr.is_v4() or addr.broadcast is None:
                continue
            out.append(addr)
    return out
